@@ -3,6 +3,8 @@
 import tensorflow as tf
 import numpy as np
 
+from scipy.sparse import issparse
+
 
 class KerasSequence(tf.keras.utils.Sequence):
     """Wraps inputs into a Keras Sequence to allow Keras models to predict on
@@ -15,6 +17,11 @@ class KerasSequence(tf.keras.utils.Sequence):
         return int(np.ceil(self.X.shape[0] / float(self.batch_size)))
     def __getitem__(self, index):
         batch_indices = range(index * self.batch_size, min((index + 1) * self.batch_size, self.X.shape[0]))
-        batch = np.array([self.X[i] for i in batch_indices])
+
+        if issparse(self.X):
+            batch = np.array([self.X[i].toarray() for i in batch_indices])
+            batch = np.squeeze(batch)
+        else:
+            batch = np.array([self.X[i] for i in batch_indices])
         placeholder_labels = self.placeholder_labels[:len(batch_indices)]
         return batch, placeholder_labels
