@@ -4,6 +4,7 @@ import json
 import os
 import shutil
 import tempfile
+from copy import copy
 import dill as pkl
 import tensorflow as tf
 import numpy as np
@@ -11,26 +12,23 @@ import numpy as np
 from tensorflow import keras
 from tensorflow.keras.callbacks import EarlyStopping
 from tensorflow.keras.models import load_model, Model
-from tensorflow.keras.layers import Dense, Input
-from tensorflow.keras import regularizers
-from sklearn.base import BaseEstimator
+from sklearn.base import BaseEstimator, TransformerMixin
+from sklearn.exceptions import NotFittedError
 
 
 from .data.generators import generator_from_neighbour_matrix, KerasSequence
-from .nn.network import triplet_network, base_network
+from .nn.network import build_supervised_layer, triplet_network, base_network
 from .nn.callbacks import ModelCheckpoint
-from .nn.losses import triplet_loss, is_categorical, is_multiclass, is_hinge
-from .nn.losses import semi_supervised_loss, validate_sparse_labels
+from .nn.losses import triplet_loss, semi_supervised_loss
 from .data.neighbour_retrieval import AnnoyKnnMatrix
-from .data.neighbour_retrieval.knn import cleanup_knn_index
 
 
-class Ivis(BaseEstimator):
+class Ivis(BaseEstimator, TransformerMixin):
     """Ivis is a technique that uses an artificial neural network for
     dimensionality reduction, often useful for the purposes of visualization.
     The network trains on triplets of data-points at a time and pulls positive
     points together, while pushing more distant points away from each other.
-    Triplets are sampled from the original data using KNN aproximation using
+    Triplets are sampled from the original data using KNN approximation using
     the Annoy library.
 
     :param int embedding_dims: Number of dimensions in the embedding space
@@ -90,9 +88,8 @@ class Ivis(BaseEstimator):
         in supervised mode. The higher the weight, the more classification
         influences training. Ignored if using Ivis in unsupervised mode.
     :param str annoy_index_path: The filepath of a pre-trained annoy index file
-        saved on disk. If provided, the annoy index file will be used.
-        Otherwise, a new index will be generated and saved to disk in the
-        current directory as 'annoy.index'.
+        saved on disk. If provided, the annoy index file will be loaded and used.
+        Otherwise, a new index will be generated and saved to disk in a temporary directory.
     :param [keras.callbacks.Callback] callbacks: List of keras Callbacks to
         pass model during training, such as the TensorBoard callback. A set of
         ivis-specific callbacks are provided in the ivis.nn.callbacks module.
@@ -144,19 +141,21 @@ class Ivis(BaseEstimator):
         self.supervised_model_ = None
         self.loss_history_ = []
         self.annoy_index_path = annoy_index_path
-
-        if callbacks is None:
-            self.callbacks = []
-        else:
-            self.callbacks = callbacks
-        for callback in self.callbacks:
-            if isinstance(callback, ModelCheckpoint):
-                callback = callback.register_ivis_model(self)
+        self.callbacks = callbacks
+        self.callbacks_ = None
         self.build_index_on_disk = build_index_on_disk
         self.neighbour_matrix = neighbour_matrix
         self.verbose = verbose
         self.n_classes = None
 
+    def _validate_parameters(self):
+        """ Validate parameters before fitting """
+        self.callbacks_ = [] if self.callbacks is None else list(map(copy, self.callbacks))
+
+        for callback in self.callbacks_:
+            if isinstance(callback, ModelCheckpoint):
+                callback.register_ivis_model(self)
+
     def __getstate__(self):
         """ Return object serializable variable dict """
 
@@ -169,6 +168,8 @@ class Ivis(BaseEstimator):
             state['supervised_model_'] = None
         if 'callbacks' in state:
             state['callbacks'] = []
+        if 'callbacks_' in state:
+            state['callbacks_'] = []
         if not isinstance(state['model'], str):
             state['model'] = None
         if 'neighbour_matrix' in state:
@@ -178,12 +179,14 @@ class Ivis(BaseEstimator):
         return state
 
     def _fit(self, X, Y=None, shuffle_mode=True):
+        self._validate_parameters()
+
         if self.neighbour_matrix is None:
             if self.annoy_index_path is None:
                 # Create a temporary folder to store the index on disk
                 temp_dir = tempfile.mkdtemp()
-                self.annoy_index_path = os.path.join(temp_dir, 'annoy.index')
-                self.neighbour_matrix = AnnoyKnnMatrix.build(X, path=self.annoy_index_path,
+                temp_index_path = os.path.join(temp_dir, 'annoy.index')
+                self.neighbour_matrix = AnnoyKnnMatrix.build(X, path=temp_index_path,
                                                              k=self.k, metric=self.knn_distance_metric,
                                                              search_k=self.search_k,
                                                              include_distances=False, ntrees=self.ntrees,
@@ -191,10 +194,11 @@ class Ivis(BaseEstimator):
                                                              precompute=self.precompute, verbose=self.verbose)
 
                 # Clean up temporary folder with index before object is garbage collected
-                weakref.finalize(self, cleanup_knn_index, self.neighbour_matrix, temp_dir)
+                weakref.finalize(self, self.neighbour_matrix.delete_index, parent=True)
             else:
                 self.neighbour_matrix = AnnoyKnnMatrix.load(self.annoy_index_path, X.shape,
-                                                            k=self.k, search_k=self.search_k,
+                                                            k=self.k, metric=self.knn_distance_metric,
+                                                            search_k=self.search_k,
                                                             include_distances=False, precompute=self.precompute,
                                                             verbose=self.verbose)
 
@@ -202,7 +206,6 @@ class Ivis(BaseEstimator):
                                                   neighbour_matrix=self.neighbour_matrix,
                                                   batch_size=self.batch_size)
 
-        loss_monitor = 'loss'
         try:
             triplet_loss_func = triplet_loss(distance=self.distance)
         except KeyError:
@@ -211,66 +214,29 @@ class Ivis(BaseEstimator):
         if self.model_ is None:
             if isinstance(self.model, str):
                 input_size = (X.shape[-1],)
-                self.model_, anchor_embedding, _, _ = \
+                self.model_, (anchor_embedding, *_) = \
                     triplet_network(base_network(self.model, input_size),
                                     embedding_dims=self.embedding_dims)
             else:
-                self.model_, anchor_embedding, _, _ = \
+                self.model_, (anchor_embedding, *_) = \
                     triplet_network(self.model,
                                     embedding_dims=self.embedding_dims)
 
             if Y is None:
                 self.model_.compile(optimizer='adam', loss=triplet_loss_func)
             else:
-                if is_categorical(self.supervision_metric):
-                    if not is_multiclass(self.supervision_metric):
-                        if not is_hinge(self.supervision_metric):
-                            # Binary logistic classifier
-                            if len(Y.shape) > 1:
-                                self.n_classes = Y.shape[-1]
-                            else:
-                                self.n_classes = 1
-                            supervised_output = Dense(self.n_classes, activation='sigmoid',
-                                                      name='supervised')(anchor_embedding)
-                        else:
-                            # Binary Linear SVM output
-                            if len(Y.shape) > 1:
-                                self.n_classes = Y.shape[-1]
-                            else:
-                                self.n_classes = 1
-                            supervised_output = Dense(self.n_classes, activation='linear',
-                                                      name='supervised',
-                                                      kernel_regularizer=regularizers.l2())(anchor_embedding)
-                    else:
-                        if not is_hinge(self.supervision_metric):
-                            validate_sparse_labels(Y)
-                            self.n_classes = len(np.unique(Y[Y != np.array(-1)]))
-                            # Softmax classifier
-                            supervised_output = Dense(self.n_classes, activation='softmax',
-                                                      name='supervised')(anchor_embedding)
-                        else:
-                            self.n_classes = len(np.unique(Y, axis=0))
-                            # Multiclass Linear SVM output
-                            supervised_output = Dense(self.n_classes, activation='linear',
-                                                      name='supervised',
-                                                      kernel_regularizer=regularizers.l2())(anchor_embedding)
-                else:
-                    # Regression
-                    if len(Y.shape) > 1:
-                        self.n_classes = Y.shape[-1]
-                    else:
-                        self.n_classes = 1
-                    supervised_output = Dense(self.n_classes, activation='linear',
-                                              name='supervised')(anchor_embedding)
+                supervised_layer = build_supervised_layer(self.supervision_metric,
+                                                          Y, name='supervised')
+                supervised_out = supervised_layer(anchor_embedding)
+                self.n_classes = supervised_layer.units
 
                 supervised_loss = keras.losses.get(self.supervision_metric)
                 if self.supervision_metric == 'sparse_categorical_crossentropy':
                     supervised_loss = semi_supervised_loss(supervised_loss)
 
-                final_network = Model(inputs=self.model_.inputs,
-                                      outputs=[self.model_.output,
-                                               supervised_output])
-                self.model_ = final_network
+                self.model_ = Model(inputs=self.model_.inputs,
+                                    outputs=[self.model_.output,
+                                             supervised_out])
                 self.model_.compile(
                     optimizer='adam',
                     loss={
@@ -282,11 +248,8 @@ class Ivis(BaseEstimator):
                         'supervised': self.supervision_weight})
 
                 # Store dedicated classification model
-                supervised_model_input = Input(shape=(X.shape[-1],))
-                embedding = self.model_.layers[3](supervised_model_input)
-                softmax_out = self.model_.layers[-1](embedding)
-
-                self.supervised_model_ = Model(supervised_model_input, softmax_out)
+                self.supervised_model_ = Model(self.model_.inputs[0],
+                                               self.model_.get_layer('supervised').output)
 
         self.encoder = self.model_.layers[3]
 
@@ -296,8 +259,8 @@ class Ivis(BaseEstimator):
         hist = self.model_.fit(
             datagen,
             epochs=self.epochs,
-            callbacks=self.callbacks + [EarlyStopping(monitor=loss_monitor,
-                                                      patience=self.n_epochs_without_progress)],
+            callbacks=self.callbacks_ + [EarlyStopping(monitor='loss',
+                                                       patience=self.n_epochs_without_progress)],
             shuffle=shuffle_mode,
             steps_per_epoch=int(np.ceil(X.shape[0] / self.batch_size)),
             verbose=self.verbose)
@@ -333,7 +296,7 @@ class Ivis(BaseEstimator):
             Data to train on and then embedded.
             Needs to have a `.shape` attribute and a `__getitem__` method.
         Y : array, shape (n_samples)
-            Optional array for supervised dimentionality reduction.
+            Optional array for supervised dimensionality reduction.
             If Y contains -1 labels, and 'sparse_categorical_crossentropy'
             is the loss function, semi-supervised learning will be used.
 
@@ -361,6 +324,10 @@ class Ivis(BaseEstimator):
             Embedding of the data in low-dimensional space.
         """
 
+
+        if self.encoder is None:
+            raise NotFittedError("Model was not fitted yet. Call `fit` before calling `transform`.")
+
         embedding = self.encoder.predict(KerasSequence(X, batch_size=self.batch_size),
                                          verbose=self.verbose)
         return embedding
@@ -383,10 +350,10 @@ class Ivis(BaseEstimator):
         """
 
         if self.supervised_model_ is None:
-            raise Exception("Model was not trained in classification mode.")
+            raise NotFittedError("Model was not trained in classification mode.")
 
-        softmax_output = self.supervised_model_.predict(X, verbose=self.verbose)
-        return softmax_output
+        supervised_output = self.supervised_model_.predict(X, verbose=self.verbose)
+        return supervised_output
 
     def save_model(self, folder_path, overwrite=False):
         """Save an ivis model
@@ -459,7 +426,7 @@ class Ivis(BaseEstimator):
                                       allow_pickle=True)
             optimizer.set_weights(optimizer_state)
 
-            self.model_, _, _, _ = triplet_network(base_model, embedding_dims=None)
+            self.model_, _ = triplet_network(base_model, embedding_dims=None)
             self.model_.compile(loss=loss_function, optimizer=optimizer)
 
         self.encoder = self.model_.layers[3]
